#### python-style classes in R

.rootPC <- ".__PC__"

# initiating the class

"PythonClass" <- function(class, extends = character(0), hash = FALSE)
{
  root <- get(".rootPC", pos=match("package:PythonClass", search()), inherits=FALSE)
  classEnvName <- paste(root, class, sep="")

  # check that extends classes exist

  notThere <- character(0)
  for (i in extends) {
    cen <- paste(root, i, sep="")
    if (!exists(cen, envir=.GlobalEnv, inherits=FALSE))
      notThere <- c(notThere, i)
  }
  if (length(notThere) > 0)
    stop(sprintf("Class(es) %s not defined\n", paste(paste("\"", notThere, "\"", sep=""),
      collapse=", ")))

  # create new class environment, and wrap in a list

  classEnv <- list(new.env())
  attr(classEnv, "extends") <- extends
  attr(classEnv, "hash") <- hash
  assign(classEnvName, classEnv, envir=.GlobalEnv)
  
  invisible(NULL)
}

# adding methods

"setMethodPC" <- function(class, name, fun)
{
  root <- get(".rootPC", pos=match("package:PythonClass", search()), inherits=FALSE)
  classEnvName <- paste(root, class, sep="")
  classEnv <- get(classEnvName, envir = .GlobalEnv)

  # check method is correctly defined using "self"

  gotSelf <- length(grep("^function \\(self[, )]", deparse(args(fun))[1])) > 0
  if (!gotSelf)
    stop("First argument of \"fun\" must be \"self\"")

  # warn if overwriting

  existingMethods <- ls(envir=classEnv[[1]], all.names=TRUE)
  if (!is.na(match(name, existingMethods)))
    cat(sprintf("PythonClass Warning: Overwriting method \"%s\" in class \"%s\"\n",
      name, class))

  # write function into the environment

  assign(name, fun, envir = classEnv[[1]])
  invisible(NULL)
}

# instantiating an object

"newPC" <- function(class, init = "init", ...)
{
  root <- get(".rootPC", pos=match("package:PythonClass", search()), inherits=FALSE)
  classEnvName <- paste(root, class, sep="")
  classEnv <- get(classEnvName, envir = .GlobalEnv)

  # create self environment (see below)

  hash <- attr(classEnv, "hash")
  self <- list(new.env(hash = hash))
  attr(self, "init") <- init
  class(self) <- "selfPC"

  # recursive enumeruation of parent classes in order

  "compileExtends" <- function(x)
  {
    if (is.null(x) || length(x) == 0) return(character(0))
    cen <- paste(root, x[1], sep="")
    ex <- attr(get(cen, envir=.GlobalEnv, inherits=FALSE), "extends")
    c(x[1], Recall(ex), Recall(x[-1]))
  }

  extends <- unique(compileExtends(class))
  attr(self, "extends") <- extends

  # set up delete list for exit

  on.exit(rm(list=setdiff(ls(all.names=TRUE), "self")))

  # create publically-available methods list

  robj <- list()
  mm <- unique(unlist(lapply(extends, function(x) {
    cen <- paste(root, x, sep="")
    ls(envir=get(cen, envir=.GlobalEnv, inherits=FALSE)[[1]]) # won't see hidden methods
  })))

  for (i in mm) {

    ss <- sprintf("self$%s", i)
    mi <- eval(parse(text = ss))

    mo <- character(2)

    # set up the method arguments

    tmp <- deparse(args(mi))
    tmp <- paste(tmp[-length(tmp)], collapse = "") # remove trailing NULL
    mo[1] <- gsub("self,? *", "", tmp)             # remove the self argument

    # set up the matching call to self$method
    
    tmp <- names(formals(mi))
    tmp <- paste(paste(tmp, "=", tmp), collapse=", ")      # echo args
    tmp <- gsub("\\.\\.\\. = \\.\\.\\.", "\\.\\.\\.", tmp) # handle ...
    mo[2] <- sprintf("%s (%s)", ss, tmp)                   # wrap into fn

    robj[[i]] <- eval(parse(text = paste(mo, collapse=" ")))
  }

  # add on a "getSelf" method

  if (is.na(match("getSelf", mm)))
    robj$getSelf <- function() self

  # and a "getSlot" method

  if (is.na(match("getSlot", mm)))
    robj$getSlot <- function(name) eval(parse(text = sprintf("self$%s", name)))

  # do initialisation

  if (is.character(init) && nchar(init) > 0) {
    if (is.na(match(init, names(robj))))
      cat(sprintf("PythonClass Warning: no method named \"%s\" for initialisation\n", init))
    else {
      robj[[init]](...)
      robj[[init]] <- NULL # for safety
    }
  } else if (length(list(...)) > 0)
    cat("PythonClass Warning: No valid \"init\" method supplied\n")

  # class and return

  class(robj) <- c(class, "PythonClass")
  robj
}

# showing the instances

#"showInstancesPC" <- function(class, envir = parent.frame())
#{
#  inst <- ls(envir=envir, all.names=TRUE)
#  target <- sapply(inst, function(x) identical(class(get(x, envir=envir)), 
#    c("PythonClass", class)))
#  robj <- inst[target]
#
#  # attach class definition attribute
#
#  root <- get(".rootPC", pos=match("package:PythonClass", search()), inherits=FALSE)
#  classEnvName <- paste(root, class, sep="")
#  attr(robj, "classEnvName") <- classEnvName
#
#  robj
#}

#### here's where we overload the $ operator but not the [[ operator

"$.selfPC" <- function(object, name)
{
  robj <- NULL
  slots <- ls(envir=object[[1]], all.names=TRUE)
  if (!is.na(match(name, slots)))
    robj <- get(name, envir=object[[1]])
  else {
    root <- get(".rootPC", pos=match("package:PythonClass", search()),
      inherits=FALSE)
    extends <- attr(object, "extends")
    for (i in extends) {
      cen <- paste(root, i, sep="")
      ce <- get(cen, envir=.GlobalEnv, inherits=FALSE)
      if (!is.na(match(name, ls(envir=ce[[1]], all.names=TRUE)))) {
        robj <- get(name, envir=ce[[1]])
        break
      }
    }
  }
  robj
}

"$<-.selfPC" <- function(object, name, value)
{
  assign(name, value, envir=object[[1]])
  object
}

#### here are some default S-style methods

"print.PythonClass" <- function(x, ...)
{
  cat(sprintf("A \"%s\" object with method(s):\n", paste(class(x), collapse=".")))
  mm <- formatC(names(x), width=10, format="s")
  Arguments <- sapply(x, function(i) {
    tmp <- deparse(args(i))
    paste(tmp[-length(tmp)], collapse = "") # remove trailing NULL
  })
  names(Arguments) <- mm
  print(data.frame(Arguments), right = FALSE, collab = "")
  invisible(Arguments)
}
