\name{PythonClass}

\alias{PythonClass}
\alias{setMethodPC}
\alias{newPC}
\alias{print.PythonClass}

\title{Object Oriented Programming using Python-Style Classes}

\description{Python-style classes are a flexible way to implement
general object-oriented concepts.  Unlike in R, in Python (like in C++
and Java) methods are bound to objects.  For example, instead of
\code{print(foo)} we would do \code{foo$print()}.  Python-style
classes are useful when we want to do complicated things with objects.
Details of this implementation are given in the Section "Python-Style
Classes".  Some implementation notes are available in the Section
"Programmer's Notes".}

\usage{
PythonClass(class, extends = character(0), hash = FALSE)
setMethodPC(class, name, fun)
newPC(class, init = "init", \dots)
\method{print}{PythonClass}(x, \dots)
}

\arguments{
  \item{class}{Character string giving the name of the class}
  \item{extends}{Character vector identifying the classes that the new
class extends}
  \item{hash}{Should the slots for instances of the class be hashed?}
  \item{name}{Character string giving the name of the method}
  \item{fun}{Function representing the method}
  \item{init}{Name of method used for initialisation}
  \item{\dots}{Arguments passed through to the \code{init} method}
  \item{x}{Object that inherits \code{PythonClass} (nothing to be passed in \dots)}
}

\details{

  \bold{Types of method.}  There are two types of method,
  public methods and hidden methods.  Both are specified
  via \code{setMethodPC}, but the latter have names that start with a
  period (\code{.}).  If \code{foo} is a \code{PythonClass} object, as
  created by \code{newPC}, then a public method \code{bar} is accessed
  using \code{foo$bar()}, or \code{foo$bar(ArgumentsToBar)}.  Inside the
  code for a given method, both types of method are available, using the
  \code{self} argument, e.g. \code{self$bar(self, ArgumentsToBar)}, or
  \code{self$.hiddenMethod(self, ArgumentsToHiddenMethod)}.

  All methods defined by
  \code{setMethodPC} must have \code{self} as their first argument, even
  trivial ones such as \code{setMethodPC("fooClass", "nullFun",
    function(self) NULL)}.   See
  Programmer's Notes below, for details about \code{self}.

  \bold{Inheritance.}  The class inherits all of the methods (including
  hidden methods) from the classes given in \code{extends}, if there are
  any.  It is an error to specify a class that has not been defined
  using \code{PythonClass}.  In class extension, methods from classes
  earlier in the \code{extends} vector overwrite the later ones.

  \bold{Access to slots.}  Each \code{PythonClass} object comes with two
  extra methods, \code{getSlot} and \code{getSelf}.  The former provides
  public access to slots.  Inside a given method, access to slots is
  done using \code{self$name}.  To assign to a slot in the current
  object, use \code{self$name <- value}.  To assign to a slot in another
  object a dereference is necessary: \code{tmp <- otherObj$getSelf();
    tmp$name <- value}.

  \bold{Instantiation.} The \code{newPC} function instantiates an object
of the specified class, and initialises it by passing further arguments
through to the method \code{init}.  This method is then deleted from the
methods that may be invoked directly, although it may still be invoked
inside a method (see above).  Rather than using \code{newPC} directly,
you may want to write an explicit "creator" function, i.e. \code{"foo"
<- function(...)  {robj <- newPC("foo", init = "fooInitFunction", ...);
robj$getSelf <- NULL; robj}}, where in this case I have also removed the
\code{getSelf} public method than comes by default.

}

\value{

\item{PythonClass}{Doesn't return anything.  It is invoked for the
side-effect of creating a new object in \code{.GlobalEnv}, which is an
environment that contains/will contain the definition of the methods
of the class.}

\item{setMethodPC}{Doesn't return anything.  It is invoked for the
side-effect of adding a method to the class environment in
\code{.GlobalEnv}.}

\item{newPC}{Returns an object of class \code{c("PythonClass",
class)}.  Objects of class \code{PythonClass} have a print method that
displays the non-hidden methods of the class.}

}

\section{Saving Python-style objects}{

Some caution has to be exercised when saving an object of class
\code{PythonClass}.  This is because these objects have a complex
structure (see below, Programmer's Notes for the whys and wherefores).
This structure is preserved using the \code{save} command, and
therefore this is the appropriate mechanism for making a file
containing a persistent copy.  Using \code{dput} or \code{dump} will
not work.  To read the object back in again, use \code{load}.  Note
that this object will not be able to find its methods unless these
have been created in the usual way using \code{PythonClass} and
\code{SetMethodPC}.

An example of saving and loading is given below.

}

\section{Bugs}{One thing to note is that lazy evaluation does not
work for public methods.}

\section{Python-Style Classes}{

The key feature of Python-style classes is the use of the \code{self}
pointer.  This is passed as the first argument of each method, and
allows the method access to all of the other methods, and also all of
the slots of the object.  In R I use the \code{$} operator to access
these.  This operator is also used to access the methods of the object
externally.

This implementation of Python-style classes offers:
\enumerate{
\item Simple and uniform access to object slots and methods, both
internally and externally.

\item Arbitrary slots that can be created on-the-fly by methods.

\item Hidden methods, that are accessible to other methods but not to
the user.

\item Ability to share objects between instances of the same class.

\item Multiple inheritance, with child methods being able to over-write
parent methods.

\item `Updatability': the effect modifying a method (by assigning a
new version using \code{setMethodPC}) of any class is seen immediately
by all instances of classes that inherit from that class.

\item `Detachability': individual methods can be detached and used as
functions.
}

}

\section{Programmer's notes}{

The methods of a given class are stored as environments in
\code{.GlobalEnv}, using the name \code{paste(.rootPC, class,
sep="")}.  If you want to create an object that is accessible to all
instances of the class then you should arrange for your method to
stick the object into this environment.  This environment also has a
\code{extends} attribute that describes the inheritance structure.

On instantiation a new object gets a \code{self} environment, which
stands in for the Python \code{self} pointer (actually implemented as
a list with attributes containing an environment as its first
component).  The \code{self} environment holds the slots of the
object.  This environment has class \code{selfPC}, allowing me to
overload the \code{$} operator.  Thus a lot of the work is done in the
functions \code{$.selfPC} and \code{$<-.selfPC}.

When a method is invoked internally it is collected from the
appropriate environment and delivered, using the \code{$} operator.
When it is invoked publically, ie by the user, a simple wrapper maps
this into an internal call with the appropriate \code{self} value.
The external representation of an object is a list of functions with a
common environment (itself containing only the \code{self}
environment), which preserves the syntax of calling a method using
\code{$}.

}

\author{Jonathan Rougier, \email{J.C.Rougier@durham.ac.uk}}

\seealso{The \code{methods} package, which extends R's existing
method-dispatch mechanism.}

\examples{
#### Here's an example of a container for an uncertain vector
#### described by a mean vector and a variance matrix

PythonClass("uuobj")

# a couple of internal methods for handling the variance

setMethodPC("uuobj", ".unpack", function(self, C) {
  C[lower.tri(C, diag=TRUE)]
})

setMethodPC("uuobj", ".pack", function(self) {
  n <- length(self$E)
  tmp <- matrix(NA, n, n)
  tmp[lower.tri(tmp, diag=TRUE)] <- self$C
  ifelse(is.na(tmp), t(tmp), tmp)
})

# initialisation

setMethodPC("uuobj", "init", function(self, E, C, nms = names(E)) {
  self$E <- as.vector(E)
  self$nms <- nms
  self$C <- self$.unpack(self, C)
  invisible(NULL)
})

# a couple of accessor methods for the mean and variance

setMethodPC("uuobj", "getE", function(self) {
  tmp <- self$E
  nms <- self$nms
  if (!is.null(nms))
    names(tmp) <- nms
  tmp
})

setMethodPC("uuobj", "getV", function(self) {
  tmp <- self$.pack(self)
  nms <- self$nms
  if (!is.null(nms))
    dimnames(tmp) <- list(nms, nms)
  tmp
})

# one for the Mahalanobis distance (NOT the recommended method!)

setMethodPC("uuobj", "mahal", function(self, x) {
  dd <- x - self$E
  drop(t(dd) \%*\% solve(self$.pack(self), dd))
})

#### instantiate an object (with some names on E)

uu <- newPC("uuobj", E = structure(1:2, names=letters[1:2]),
  C = matrix(c(1, 1, 1, 2), 2, 2))

uu # use default PythonClass print

uu$getE()
uu$getV()
uu$mahal(c(0.5, 0.5))

# check the slots

uu$getSlot("C")
ls(envir = uu$getSelf()[[1]]) # more ambitious

# detach a method and use separately

foo <- uu$mahal
foo(c(1, 1)); foo(c(2, 2))

# update the mahal method (still not the best method!)

setMethodPC("uuobj", "mahal", function(self, x) {
  cat("New method being used\n")
  ch <- self$chol
  if (is.null(ch)) {
    cat("Doing the choleski decomposition\n")
    ch <- chol(self$.pack(self))
    self$chol <- ch
  }
  dd <- x - self$E
  drop(t(dd) \%*\% chol2inv(ch) \%*\% dd)
})

uu$mahal(c(2, 2))  # both have picked up the new method
foo(c(2, 2))       # and chol is only called once

# making persistent copies of an object

tf <- tempfile()
save(list = "uu", file = tf)
rm(list = "uu")

load(tf)
uu$getE() # there it is again!

unlink(tf)

}

\keyword{programming}
\keyword{classes}
\keyword{methods}
